;; prolog.el --- major mode for editing and running prolog code

;;; Copyright (C) 1986, 1987, 1997 Free Software Foundation, Inc.
;;
;; Author: Emil Astrom <emil@sics.se>
;;         Milan Zamazal <pdm@fi.muni.cz>
;;         (see below for more details)
;; Keywords: prolog major mode sicstus swi
;; Version: 0.1 (Oct 24 1997)
;;
;; Original author: Masanobu UMEDA <umerin@flab.flab.fujitsu.junet>
;; Parts of this file was taken from a modified version of the original
;; by Johan Andersson, Peter Olin, Mats Carlsson, Johan Bevemyr, Stefan
;; Andersson, and Per Danielsson (all SICS people), and Henrik Bkman
;; at Uppsala University, Sweden.
;;
;; Some ideas and also a few lines of code have been borrowed (not stolen ;-)
;; from Oz.el, the Emacs major mode for the Oz programming language,
;; Copyright (C) 1993 DFKI GmbH, Germany, with permission.
;; Authors: Ralf Scheidhauer and Michael Mehl ([scheidhr|mehl]@dfki.uni-sb.de)

;;; Commentary:
;;
;; This package provides a major mode for editing Prolog.  It knows
;; about Prolog syntax and comments, and can send regions to an inferior
;; Prolog interpreter process. Font-lock and auto indentation is
;; supported.
;;
;; This version of the mode is distributed with SICStus Prolog 3#6, but
;; please note that this is something of a pre-release, we aim to
;; extend the functionality before we release version 1.0.
;;
;; The code requires the comint, easymenu, info and font-lock libraries.
;; These are normally distributed with GNU Emacs.

;;; Installation:
;;
;; Insert the following lines in your ~/.emacs (or _emacs in case
;; of Windows/DOS) to make Emacs use this mode when editing Prolog files:
;;
;; (setq load-path (cons "/usr/local/lib/sicstus3" load-path))
;; (autoload 'run-prolog "prolog" "Start a Prolog sub-process." t)
;; (autoload 'prolog-mode "prolog" "Major mode for editing Prolog programs." t)
;; (setq prolog-system 'sicstus)
;; (setq auto-mode-alist (cons '("\\.pl$" . prolog-mode) auto-mode-alist))
;;
;; where the path in the first line is the file system path to this file.
;; MSDOS paths can be written like "d:/programs/emacs-19.34/site-lisp". 
;;
;; The last line above makes sure that files ending with .pl are
;; assumed to be Prolog files and not Perl, which is the default Emacs
;; setting. If this is not wanted, remove that line. It is then necessary
;; for the user to manually switch to prolog mode by typing
;; M-x prolog-mode after opening a prolog file.
;;
;; If the shell command to start the prolog process ('sicstus' or 'pl' for SWI
;; prolog) is not available in the default path, then it is necessary to
;; set the value of the environment variable EPROLOG to a shell command
;; to invoke the prolog process.

;;; Known problems:
;;
;; o Quoted atoms are not strings in the syntax table. This means that
;;   minor modes such as the show-paren-mode will sometimes behave
;;   strangely since they match on characters inside atoms.

;;; Future work:
;;
;; o prolog-insert-predicate-name could also insert commas and parens?
;; o Caching of neutral points to speed up parsing (and thus indentation)
;; o If syntax error in SICStus, then we could parse the code to find
;;   the exact location.


;;; Code:

(require 'comint)
(require 'easymenu)

;;-------------------------------------------------------------------
;; User configurable variables
;;-------------------------------------------------------------------

;; General configuration

(defvar prolog-system 'sicstus
  "*Prolog interpreter/compiler used.
The value of this variable is nil or a symbol.
If it is a symbol, it determines default values of other configuration
variables with respect to properties of the specified Prolog
interpreter/compiler.

Currently recognized symbol values are:
sicstus - SICStus Prolog
swi - SWI Prolog")

;; Indentation

(defvar prolog-indent-width tab-width
  "*The indentation width used by the editing buffer.")

(defvar prolog-align-comments-flag t
  "*Non-nil means automatically align comments when indenting.")

(defvar prolog-indent-mline-comments-flag t
  "*If non-nil, indent contents of /* */ comments.
Otherwise leave such lines as they are.")

(defvar prolog-object-end-to-0-flag t
  "*If non-nil, indent closing '}' in object definitions to level 0.
Otherwise indent to `prolog-indent-width'.")

(defvar prolog-left-indent-regexp "\\(;\\|\\*?->\\)"
  "*Regexp for character sequences after which next line is indented.
Next line after such a regexp is indented to the opening paranthesis level.")

(defvar prolog-paren-indent 4
  "*The indentation increase for parenthesis expressions.
Only used in ( If -> Then ; Else) and ( Disj1 ; Disj2 ) style expressions.")

;; Font locking

(defvar prolog-keywords
  (cond
   ((eq prolog-system 'sicstus)
    '("block" "dynamic" "mode" "module" "multifile" "meta_predicate"
      "parallel" "public" "sequential"))
   ((eq prolog-system 'swi)
    '("discontiguous" "dynamic" "ensure_loaded" "export" "export_list" "import"
      "meta_predicate" "module" "module_transparent" "multifile" "require"
      "use_module" "volatile"))
   (t
    '("dynamic" "module")))
  "*A list of prolog keywords which is used for font-locking of directives.")

;; Keyboard

(defvar prolog-electric-newline-flag t
  "*If non-nil, automatically indent the next line when the user types RET.")

(defvar prolog-hungry-delete-key-flag nil
  "*If non-nil, delete key consumes all preceding spaces.")

(defvar prolog-old-sicstus-keys-flag nil
  "*If non-nil, old SICStus Prolog mode keybindings are used.")

;; Inferior mode

(defvar prolog-program-name
		  (or (getenv "EPROLOG")
		      (cond
		       ((eq prolog-system 'sicstus) "sicstus")
		       ((eq prolog-system 'swi) "pl")
		       (t "prolog")))
  "*Program name for invoking an inferior Prolog with `run-prolog'.")

(defvar prolog-program-switches
		  (cond
		   ((eq prolog-system 'sicstus) '("-i"))
		   (t nil))
  "*List of switches given to inferior Prolog run with `run-prolog'.")

(defvar prolog-consult-string
		  (cond
		   ((eq prolog-system 'sicstus)
		    "prolog:zap_file(%m,%b,consult).")
		   ((eq prolog-system 'swi) "[%f].")
		   (t "reconsult(%f)."))
  "*String defining predicate for reconsulting.
Some parts of the string are replaced:
`%f' by the name of the consulted file (can be a temporary file)
`%b' by the file name of the buffer to consult
`%m' by the module name and name of the consulted file separated by colon")

(defvar prolog-compile-string
		  (cond
		   ((eq prolog-system 'sicstus)
		    "prolog:zap_file(%m,%b,compile).")
		   ((eq prolog-system 'swi) "[%f].")
		   (t "compile(%f)."))
  "*String defining predicate for recompilation.
Some parts of the string are replaced:
`%f' by the name of the compiled file (can be a temporary file)
`%b' by the file name of the buffer to compile
`%m' by the module name and name of the compiled file separated by colon")

(defvar prolog-eof-string "end_of_file.\n"
  "*String that represents end of file for prolog.
nil means send actual operating system end of file.")

(defvar prolog-prompt-regexp
  (cond
   ((eq prolog-system 'sicstus) "^| [ ?][- ] *")
   ((eq prolog-system 'swi) "^[1-9][0-9]* \\?- \\|^| +")
   (t "^ *\\?-"))
  "*Prompt of the prolog system command line.")

(defvar prolog-continued-prompt-regexp
  (cond
   ((eq prolog-system 'sicstus) "^\\(| +\\|     +\\)")
   (t "^|: +"))
  "*Regexp matching the prompt when consulting `user'.")

(defvar prolog-debug-on-string "debug.\n"
  "*Predicate for enabling debug mode.")

(defvar prolog-debug-off-string "nodebug.\n"
  "*Predicate for disabling debug mode.")

(defvar prolog-trace-on-string "trace.\n"
  "*Predicate for enabling tracing.")

(defvar prolog-trace-off-string "notrace.\n"
  "*Predicate for disabling tracing.")

(defvar prolog-use-standard-consult-compile-method-flag t
  "*If non-nil use the standard compilation method.
Otherwise the new compilation method will be used. This
utilises a special compilation buffer with the associated
features such as parsing of error messages and automatically
jumping to the source code responsible for the error.

Warning: the new method is so far only experimental and
does contain bugs. The recommended setting for the novice user
is non-nil for this variable.")


;; Miscellaneous

(defvar prolog-use-prolog-tokenizer-flag t
  "*If non-nil, then use the internal prolog tokenizer for indentation etc.
Otherwise use `parse-partial-sexp' which is faster but sometimes incorrect.")



;;-------------------------------------------------------------------
;; The user can mess with these at his/her own risk!
;;-------------------------------------------------------------------

(defvar prolog-mode-syntax-table nil)
(defvar prolog-mode-abbrev-table nil)
(defvar prolog-mode-map nil)
  
(defconst prolog-left-paren "[[({]" 
  "The characters that are used as left parentheses for the indentation code.")
(defconst prolog-right-paren "[])}]"
  "The characters that are used as right parentheses for the indentation code.")
(defconst prolog-atom-char-regexp "[a-zA-Z0-9_$]"
  "Regexp specifying characters which constitute atoms without quoting.")
(defconst prolog-atom-regexp (format "[a-z$]+%s*" prolog-atom-char-regexp)
  "Regexp specifying non-quoted atoms.")
(defconst prolog-quoted-atom-regexp
  "[^0-9]\\('\\([^\n']\\|\\\\'\\)*[^\\]'\\)"
  "Regexp matching a quoted atom.")
(defconst prolog-string-regexp
  "\\(\"\\([^\n\"]\\|\\\\\"\\)*[^\\]\"\\)"
  "Regexp matching a string.")
(defconst prolog-head-delimiter "\\(:-\\|\\+:\\|-:\\|\\+\\?\\|-\\?\\|-->\\)"
  "A regexp for matching on the end delimiter of a head (e.g. \":-\").")

(defvar prolog-compilation-buffer "*prolog-compilation*"
  "Name of the output buffer for Prolog compilation/consulting.")


;;-------------------------------------------------------------------
;; Prolog mode
;;-------------------------------------------------------------------

(if prolog-mode-syntax-table
    ()
  (let ((table (make-syntax-table)))
    (modify-syntax-entry ?_ "_" table)
    (modify-syntax-entry ?* ". 23b" table)
    (modify-syntax-entry ?/ ". 124" table)
    (modify-syntax-entry ?+ "." table)
    (modify-syntax-entry ?- "." table)
    (modify-syntax-entry ?= "." table)
    (modify-syntax-entry ?% "<" table)
    (modify-syntax-entry ?\n ">" table)
    (modify-syntax-entry ?< "." table)
    (modify-syntax-entry ?> "." table)
    (modify-syntax-entry ?| "." table)
    (modify-syntax-entry ?\' "." table)
    (setq prolog-mode-syntax-table table)))

(define-abbrev-table 'prolog-mode-abbrev-table ())

(defun prolog-mode-variables ()
  "Set some common variables to Prolog code specific values."
  (setq local-abbrev-table prolog-mode-abbrev-table)
  (make-local-variable 'paragraph-start)
  (setq paragraph-start (concat "%%\\|$\\|" page-delimiter)) ;'%%..'
  (make-local-variable 'paragraph-separate)
  (setq paragraph-separate paragraph-start)
  (make-local-variable 'paragraph-ignore-fill-prefix)
  (setq paragraph-ignore-fill-prefix t)
  (make-local-variable 'adaptive-fill-mode)
  (setq adaptive-fill-mode nil)
  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'prolog-indent-line)
  (make-local-variable 'comment-start)
  (setq comment-start "%")
  (make-local-variable 'comment-end)
  (setq comment-end "")
  (make-local-variable 'comment-start-skip)
  ;; This complex regexp makes sure that comments cannot start
  ;; inside quoted atoms or strings
  (setq comment-start-skip 
	(format "^\\(\\(%s\\|%s\\|[^\n\\'\\\"%%]\\)*\\)\\(/\\*+ *\\|%%+ *\\)" 
		prolog-quoted-atom-regexp prolog-string-regexp))
  (make-local-variable 'comment-column)
  (make-local-variable 'comment-indent-function)
  (setq comment-indent-function 'prolog-comment-indent)
  (make-local-variable 'comment-indent-function)
  (setq comment-indent-function 'prolog-comment-indent)
)

(defun prolog-mode-keybindings-common (map)
  "Define keybindings common to both Prolog modes."
  (define-key map "\C-c?" 'prolog-help-on-predicate)
  (define-key map "\C-c/" 'prolog-help-apropos)
  (define-key map "\C-c\C-d" 'prolog-debug-on)
  (define-key map "\C-c\C-t" 'prolog-trace-on)
  (define-key map "\C-c\r" 'run-prolog))

(defun prolog-mode-keybindings-edit (map)
  "Define keybindings for prolog-mode."
  (define-key map "\M-{" 'prolog-goto-clause-start)
  (define-key map "\M-}" 'prolog-goto-clause-end)
  (define-key map "\M-\C-c" 'prolog-mark-clause)
  (define-key map "\C-\M-a" 'prolog-goto-pred-start)
  (define-key map "\C-\M-e" 'prolog-goto-pred-end)
  (define-key map "\M-\C-h" 'prolog-mark-predicate)
  (define-key map "\M-\C-n" 'prolog-forward-list)
  (define-key map "\M-\C-p" 'prolog-backward-list)
  (define-key map "\M-p" 'prolog-insert-predicate-name)
  (define-key map "\M-\r" 'prolog-insert-next-clause)

  (define-key map "\177" 'prolog-electric-delete)
  (if prolog-electric-newline-flag 
      (define-key map "\r" 'prolog-electric-newline))

  (if prolog-old-sicstus-keys-flag
      (progn
	(define-key map "\C-c\C-c" 'prolog-consult-predicate)
	(define-key map "\C-cc" 'prolog-consult-region)
	(define-key map "\C-cC" 'prolog-consult-buffer)
	(define-key map "\C-c\C-k" 'prolog-compile-predicate)
	(define-key map "\C-ck" 'prolog-compile-region)
	(define-key map "\C-cK" 'prolog-compile-buffer))
    (define-key map "\C-c\C-p" 'prolog-consult-predicate)
    (define-key map "\C-c\C-r" 'prolog-consult-region)
    (define-key map "\C-c\C-b" 'prolog-consult-buffer)
    (define-key map "\C-c\C-cp" 'prolog-compile-predicate)
    (define-key map "\C-c\C-cr" 'prolog-compile-region)
    (define-key map "\C-c\C-cb" 'prolog-compile-buffer)))

(defun prolog-mode-keybindings-inferior (map)
  "Define keybindings for inferior-prolog-mode."
  ;; No inferior mode specific keybindings now.
  )

(if prolog-mode-map
    ()
  (setq prolog-mode-map (make-sparse-keymap))
  (prolog-mode-keybindings-common prolog-mode-map)
  (prolog-mode-keybindings-edit prolog-mode-map))

;;;###autoload
(defun prolog-mode ()
  "Major mode for editing Prolog code.

Blank lines and `%%...' separate paragraphs.  `%'s starts a comment
line and comments can also be enclosed in /* ... */.

Commands:
\\{prolog-mode-map}
Entry to this mode calls the value of `prolog-mode-hook'
if that value is non-nil."
  (interactive)
  (kill-all-local-variables)
  (use-local-map prolog-mode-map)
  (setq major-mode 'prolog-mode)
  (setq mode-name (concat "Prolog"
			  (cond
			   ((eq prolog-system 'sicstus) "[SICStus]")
			   ((eq prolog-system 'swi) "[SWI]")
			   (t ""))))
  (set-syntax-table prolog-mode-syntax-table)
  (prolog-mode-variables)
  (run-hooks 'prolog-mode-hook))


;;-------------------------------------------------------------------
;; Inferior prolog mode
;;-------------------------------------------------------------------

(defvar inferior-prolog-mode-map nil)
(defvar inferior-prolog-mode-hook nil)

(defun inferior-prolog-mode ()
  "Major mode for interacting with an inferior Prolog process.

The following commands are available:
\\{inferior-prolog-mode-map}

Entry to this mode calls the value of prolog-mode-hook with no arguments,
if that value is non-nil.  Likewise with the value of comint-mode-hook.
prolog-mode-hook is called after comint-mode-hook.

You can send text to the inferior Prolog from other buffers
using the commands send-region, send-string and \\[prolog-consult-region].

Commands:
Tab indents for Prolog; with argument, shifts rest
 of expression rigidly with the current line.
Paragraphs are separated only by blank lines and '%%'. '%'s start comments.

Return at end of buffer sends line as input.
Return not at end copies rest of line to end and sends it.
\\[comint-delchar-or-maybe-eof] sends end-of-file as input.
\\[comint-kill-input] and \\[backward-kill-word] are kill commands, imitating normal Unix input editing.
\\[comint-interrupt-subjob] interrupts the shell or its current subjob if any.
\\[comint-stop-subjob] stops, likewise. \\[comint-quit-subjob] sends quit signal, likewise."
  (interactive)
  (cond ((not (eq major-mode 'inferior-prolog-mode))
	 (kill-all-local-variables)
	 (comint-mode)
	 (setq comint-input-filter 'prolog-input-filter)
	 (setq major-mode 'inferior-prolog-mode)
	 (setq mode-name "Inferior Prolog")
	 (setq mode-line-process '(": %s"))
	 (prolog-mode-variables)
	 (if inferior-prolog-mode-map
	     ()
	   (setq inferior-prolog-mode-map (copy-keymap comint-mode-map))
	   (prolog-mode-keybindings-common inferior-prolog-mode-map)
	   (prolog-mode-keybindings-inferior inferior-prolog-mode-map))
	 (use-local-map inferior-prolog-mode-map)
	 (setq comint-prompt-regexp prolog-prompt-regexp)
	 (make-variable-buffer-local 'shell-dirstack-query)
	 (setq shell-dirstack-query "pwd.")
	 (run-hooks 'inferior-prolog-mode-hook))))

(defun prolog-input-filter (str)
  (cond ((string-match "\\`\\s *\\'" str) nil) ;whitespace
	((not (eq major-mode 'inferior-prolog-mode)) t)
	((= (length str) 1) nil)	;one character
	((string-match "\\`[rf] *[0-9]*\\'" str) nil) ;r(edo) or f(ail)
	(t t)))

;;;###autoload
(defun run-prolog (prefix)
  "Run an inferior Prolog process, input and output via buffer *prolog*.
With prefix argument, restart the Prolog process if running before."
  (interactive "P")
  (if (and prefix (get-process "prolog"))
      (progn
	(process-send-string "prolog" "halt.\n")
	(while (get-process "prolog") (sit-for 0.1))))
  (let ((buff (buffer-name)))
    (if (not (string= buff "*prolog*"))
	(switch-to-buffer-other-window "*prolog*"))
    (apply 'make-comint "prolog" prolog-program-name nil
	   prolog-program-switches)
    (inferior-prolog-mode)))

(defun ensure-prolog-process (&optional wait)
  "If Prolog process is not running, run it.
If the optional argument WAIT is non-nil, wait for Prolog prompt specified by
the variable `prolog-prompt-regexp'."
  (if (comint-check-proc "*prolog*")
      ()
    (apply 'make-comint "prolog" prolog-program-name nil
	   prolog-program-switches)
    (save-excursion
      (set-buffer "*prolog*")
      (inferior-prolog-mode)
      (if wait
	  (progn
	    (goto-char (point-max))
	    (while
		(save-excursion
		  (not
		   (re-search-backward
		    (concat "\\(" prolog-prompt-regexp "\\)" "\\=")
		    nil t)))
	      (sit-for 0.1)))))))

(defun prolog-process-insert-string (process string)
  "Insert STRING into inferior Prolog buffer running PROCESS."
  ;; Copied from elisp manual, greek to me
  (let ((buf (current-buffer)))
    (unwind-protect
	(let (moving)
	  (set-buffer (process-buffer process))
	  (setq moving (= (point) (process-mark process)))
	  (save-excursion
	    ;; Insert the text, moving the process-marker.
	    (goto-char (process-mark process))
	    (insert string)
	    (set-marker (process-mark process) (point)))
	  (if moving (goto-char (process-mark process))))
      (set-buffer buf))))


;;------------------------------------------------------------
;; Old consulting and compiling functions
;;------------------------------------------------------------

(defun prolog-old-process-region ( compilep start end)
  "Process the region limited by START and END positions.
If COMPILEP is non-nil then use compilation, otherwise consulting."
   (interactive "r")
   (ensure-prolog-process)
   (let ((tmpfile (prolog-temporary-file))
	 (process (get-process "prolog")))
     (write-region start end tmpfile)
     (process-send-string
      "prolog" (prolog-build-prolog-command
		compilep tmpfile (prolog-bsts buffer-file-name)))
;     (delete-file tmpfile)
     (switch-to-buffer-other-window "*prolog*")))

(defun prolog-old-process-buffer ( compilep )
  "Process the entire buffer.
If COMPILEP is non-nil then use compilation, otherwise consulting."
  (interactive)
  (save-some-buffers)
  (ensure-prolog-process)
  (let ((filename (prolog-bsts buffer-file-name)))
    (process-send-string
     "prolog" (prolog-build-prolog-command 
	       compilep filename filename))
    (switch-to-buffer-other-window "*prolog*")))

(defun prolog-old-process-predicate ( compilep )
  "Process the predicate around point.
If COMPILEP is non-nil then use compilation, otherwise consulting."
  (interactive)
  (prolog-old-process-region
   compilep (prolog-pred-start) (prolog-pred-end)))


;;------------------------------------------------------------
;; Consulting and compiling
;;------------------------------------------------------------

;;; This is experimental code!

(defun prolog-temporary-file ()
  "Make temporary file name for compilation."
  (make-temp-name 
   (concat 
    (or
     (getenv "TEMP") 
     (getenv "TMP")
     (getenv "SYSTEMP")
     (getenv "TMPDIR")
     "/tmp")
    "/prolcomp")))

(defun prolog-build-prolog-command (compilep file buffername)
  "Make Prolog command for FILE compilation/consulting.
If COMPILEP is non-nil, consider compilation, otherwise consulting."
  (let* ((compile-string
	  (if compilep prolog-compile-string prolog-consult-string))
	 (module (prolog-buffer-module))
	 (file-name (concat "'" file "'"))
	 (module-name (if module (concat "'" module "'")))
	 (module-file (if module
			  (concat module-name ":" file-name)
			file-name))
	 strbeg strend)
    (if (not (string-match "^'.*'$" buffername))
	(setq buffername (concat "'" buffername "'")))
    (while (string-match "%m" compile-string)
      (setq strbeg (substring compile-string 0 (match-beginning 0)))
      (setq strend (substring compile-string (match-end 0)))
      (setq compile-string (concat strbeg module-file strend)))
    (while (string-match "%f" compile-string)
      (setq strbeg (substring compile-string 0 (match-beginning 0)))
      (setq strend (substring compile-string (match-end 0)))
      (setq compile-string (concat strbeg file-name strend)))
    (while (string-match "%b" compile-string)
      (setq strbeg (substring compile-string 0 (match-beginning 0)))
      (setq strend (substring compile-string (match-end 0)))
      (setq compile-string (concat strbeg buffername strend)))
    (concat compile-string "\n")))

(defun prolog-buffer-module ()
  "Select Prolog module name appropriate for current buffer.
Bases decision on buffer contents (-*- line)."
  ;; Look for -*- ... module: MODULENAME; ... -*-
  (let (beg end)
    (save-excursion
      (goto-char (point-min))
      (skip-chars-forward " \t")
      (and (search-forward "-*-" (save-excursion (end-of-line) (point)) t)
	   (progn
	     (skip-chars-forward " \t")
	     (setq beg (point))
	     (search-forward "-*-" (save-excursion (end-of-line) (point)) t))
	   (progn
	     (forward-char -3)
	     (skip-chars-backward " \t")
	     (setq end (point))
	     (goto-char beg)
	     (and (let ((case-fold-search t))
		    (search-forward "module:" end t))
		  (progn
		    (skip-chars-forward " \t")
		    (setq beg (point))
		    (if (search-forward ";" end t)
			(forward-char -1)
		      (goto-char end))
		    (skip-chars-backward " \t")
		    (buffer-substring beg (point)))))))))

;; Global variables for process filter function
(defvar prolog-process-flag nil
  "Non-nil means that a prolog task (i.e. a consultation or compilation job) 
is running.")
(defvar prolog-consult-compile-output ""
  "Holds the unprocessed output from the current prolog task.")
(defvar prolog-consult-compile-first-line 1
  "The number of the first line of the file to consult/compile. Used for
temporary files.")
(defvar prolog-consult-compile-file nil
  "The file to compile/consult (can be a temporary file).")
(defvar prolog-consult-compile-real-file nil
  "The file name of the buffer to compile/consult.")

(defun prolog-consult-compile (compilep file &optional first-line)
  "Consult/compile FILE.
If COMPILEP is non-nil, perform compilation, otherwise perform CONSULTING.
COMMAND is a string described by the variables `prolog-consult-string'
and `prolog-compile-string'.
Optional argument FIRST-LINE is the number of the first line in the compiled
region.

This function must be called from the source code buffer."
  (if prolog-process-flag
      (error "Another Prolog task is running."))
  (ensure-prolog-process t)
  (let* ((buffer (get-buffer-create prolog-compilation-buffer))
	 (real-file buffer-file-name)
	 (command-string (prolog-build-prolog-command compilep file real-file))
	 (process (get-process "prolog"))
	 (old-filter (process-filter process)))
    (save-excursion
      (set-buffer buffer)
      (delete-region (point-min) (point-max))
      (compilation-mode)
      (if window-system
	  (progn
	    (make-local-variable 'compilation-mode-font-lock-keywords)
	    (setq compilation-mode-font-lock-keywords
		  prolog-compilation-font-lock-keywords)))
      (if (eq prolog-system 'sicstus)
	  (progn
	    (make-local-variable 'compilation-parse-errors-function)
	    (setq compilation-parse-errors-function
		  'prolog-parse-sicstus-compilation-errors)))
      (insert command-string "\n"))
    (save-selected-window
      (pop-to-buffer buffer))
    (setq prolog-process-flag t
	  prolog-consult-compile-output ""
	  prolog-consult-compile-first-line (if first-line (1- first-line) 0)
	  prolog-consult-compile-file file
	  prolog-consult-compile-real-file (if (string= 
						file buffer-file-name)
					       nil
					     real-file))
    (save-excursion
      (set-buffer buffer)
      (goto-char (point-max))
      (set-process-filter process 'prolog-consult-compile-filter)
      (process-send-string "prolog"
			   (prolog-build-prolog-command compilep file real-file))
      (while (and prolog-process-flag
		  (accept-process-output process 10)) ; 10 secs is ok?
	(sit-for 0.1)
	(if (not (get-process "prolog"))
	    (setq prolog-process-flag nil)))
      (insert (if compilep
		  "\nCompilation finished.\n"
		"\nConsulted.\n"))
      (set-process-filter process old-filter))))

(defun prolog-parse-sicstus-compilation-errors (limit atleast)
  "Parse the prolog compilation buffer for errors.
For use with the `compilation-parse-errors-function' variable."
  (setq compilation-error-list nil)
  (message "Parsing SICStus error messages...")
  (let (filepath dir file errorline)
    (while 
	(re-search-backward
	 "{\\([a-zA-Z ]* ERROR\\|Warning\\):.* in line[s ]*\\([0-9]+\\)"
	 limit t)
      (setq errorline (string-to-number (match-string 2)))
      (save-excursion
	(re-search-backward
	 "{\\(consulting\\|compiling\\|processing\\) \\(.*\\)\\.\\.\\.}" limit t)

	(setq filepath (match-string 2)))

;;; Does this work with SICStus under Windows (i.e. backslahes and stuff?)
      (if (string-match "\\(.*/\\)\\([^/]*\\)$" filepath)
	  (progn
	    (setq dir (match-string 1 filepath))
	    (setq file (match-string 2 filepath))))

      (setq compilation-error-list
	    (cons
	     (cons (save-excursion
		     (beginning-of-line)
		     (point-marker))
		   (list (list file dir) errorline))
	     compilation-error-list)
	    ))
    ))

(defun prolog-consult-compile-filter (process output)
  "Filter function for Prolog compilation process."
  ;;(message "start")
  (setq prolog-consult-compile-output
	(concat prolog-consult-compile-output output))
  ;;(message "pccf1: %s" prolog-consult-compile-output)
  ;; Iterate through the lines of prolog-consult-compile-output
  (let (outputtype)
    (while (and prolog-process-flag
		(or
		 ;; Trace question
		 (progn 
		   (setq outputtype 'trace)
		   (and (eq prolog-system 'sicstus)
			(string-match "^[ \t]*[0-9]+[ \t]*[0-9]+[ \t]*Call:.*? "
				      prolog-consult-compile-output)))
		 
		 ;; Match anything
		 (progn 
		   (setq outputtype 'normal)
		   (string-match "^.*\n" prolog-consult-compile-output))
		   ))
      ;;(message "outputtype: %s" outputtype)

      (setq output (match-string 0 prolog-consult-compile-output))
      ;; remove the text in output from prolog-consult-compile-output
      (setq prolog-consult-compile-output
	    (substring prolog-consult-compile-output (length output)))
      ;;(message "pccf2: %s" prolog-consult-compile-output)
      
      ;; If temporary files were used, then we change the error
      ;; messages to point to the original source file.
      (cond

       ;; If the prolog process was in trace mode then it requires
       ;; user input
       ((and (eq prolog-system 'sicstus) 
	     (eq outputtype 'trace))
	(let (input)
	  (setq input (concat (read-string output) "\n"))
	  (process-send-string "prolog" input)
	  (setq output (concat output input))))

       ((eq prolog-system 'sicstus)
	(if (and prolog-consult-compile-real-file
		 (string-match
		  "\\({.*:.* in line[s ]*\\)\\([0-9]+\\)-\\([0-9]+\\)" output))
	    (setq output (replace-match
			  ;; Adds a {processing ...} line so that 
			  ;; `prolog-parse-sicstus-compilation-errors'
			  ;; finds the real file instead of the temporary one.
			  ;; Also fixes the line numbers.
			  (format "Added by Emacs: {processing %s...}\n%s%d-%d"
				  prolog-consult-compile-real-file
				  (match-string 1 output)
				  (+ prolog-consult-compile-first-line
				     (string-to-number
				      (match-string 2 output)))
				  (+ prolog-consult-compile-first-line
				     (string-to-number
				      (match-string 3 output))))
			  t t output)))
	)
       
       ((eq prolog-system 'swi)
	(if (and prolog-consult-compile-real-file
		 (string-match (format
				"%s\\([ \t]*:[ \t]*\\)\\([0-9]+\\)"
				prolog-consult-compile-file)
			       output))
	    (setq output (replace-match
			  ;; Real filename + text + fixed linenum
			  (format "%s%s%d"
				  prolog-consult-compile-real-file
				  (match-string 1 output)
				  (+ prolog-consult-compile-first-line
				     (string-to-number
				      (match-string 2 output))))
			  t t output)))
	)
       
       (t ())
       )
      ;; Write the output in the *prolog-compilation* buffer
      (insert output)))

  ;; If the prompt is visible, then the task is finished
  (if (string-match prolog-prompt-regexp prolog-consult-compile-output)
      (setq prolog-process-flag nil)))

(defun prolog-consult-compile-buffer (compilep)
  "Consult/compile current buffer.
If COMPILEP is non-nil, compile, otherwise consult."
  (let ((file buffer-file-name))
    (if file
	(progn
	  (save-some-buffers)
	  (prolog-consult-compile compilep file))
      (prolog-consult-compile-region compilep (point-min) (point-max)))))

(defun prolog-consult-compile-region (compilep beg end)
  "Consult/compile region.
If COMPILEP is non-nil, compile, otherwise consult."
  (let ((file (prolog-temporary-file))
	(lines (count-lines 1 beg)))
    (write-region beg end file nil 'no-message)
    (write-region "\n" nil file t 'no-message)
    (prolog-consult-compile compilep file
			    (if (looking-at "^") (1+ lines) lines))
    (delete-file file)))

(defun prolog-consult-compile-predicate (compilep)
  "Consult/compile the predicate around current point.
If COMPILEP is non-nil, compile, otherwise consult."
  (prolog-consult-compile-region
   compilep (prolog-pred-start) (prolog-pred-end)))

;;; Interactive interface functions, used by both the standard
;;; and the experimental consultation and compilation functions
(defun prolog-consult-buffer ()
  "Consult buffer."
  (interactive)
  (if prolog-use-standard-consult-compile-method-flag
      (prolog-old-process-buffer nil)
    (prolog-consult-compile-buffer nil)))

(defun prolog-consult-region (beg end)
  "Consult region."
  (interactive "r")
  (if prolog-use-standard-consult-compile-method-flag
      (prolog-old-process-region nil beg end)
    (prolog-consult-compile-region nil beg end)))

(defun prolog-consult-predicate ()
  "Consult the predicate around current point."
  (interactive)
  (if prolog-use-standard-consult-compile-method-flag
      (prolog-old-process-predicate nil)
    (prolog-consult-compile-predicate nil)))

(defun prolog-compile-buffer ()
  "Compile buffer."
  (interactive)
  (if prolog-use-standard-consult-compile-method-flag
      (prolog-old-process-buffer t)
    (prolog-consult-compile-buffer t)))

(defun prolog-compile-region (beg end)
  "Compile region."
  (interactive "r")
  (if prolog-use-standard-consult-compile-method-flag
      (prolog-old-process-region t beg end)
    (prolog-consult-compile-region t beg end)))

(defun prolog-compile-predicate ()
  "Compile the predicate around current point."
  (interactive)
  (if prolog-use-standard-consult-compile-method-flag
      (prolog-old-process-predicate t)
    (prolog-consult-compile-predicate t)))


;;-------------------------------------------------------------------
;; Font-lock stuff
;;-------------------------------------------------------------------

;; Auxilliary functions
(defun prolog-make-keywords-regexp (keywords)
  "Create regexp from the list of strings."
  (concat (mapconcat 'identity keywords  "\\|")))

(defun prolog-font-lock-object-matcher (bound)
  "Find SICStus objects method name for font lock."
  (let (point
	 (case-fold-search nil))
    (while (and (not point)
		(re-search-forward "\\(::[ \t\n]*{\\|&\\)[ \t]*"
				   bound t))
      (while (or (re-search-forward "\\=\n[ \t]*" bound t)
		  (re-search-forward "\\=%.*" bound t)
		  (and (re-search-forward "\\=/\\*" bound t)
		       (re-search-forward "\\*/[ \t]*" bound t))))
      (setq point (re-search-forward
		   (format "\\=\\(%s\\)" prolog-atom-regexp)
		   bound t)))
    point))

;; Set everything up
(if window-system 
    (progn
      (require 'font-lock)
      ;; This is only for backward compatibility. Not needed with
      ;; Emacs 19.34 or later.
      (if (boundp 'font-lock-background-mode)
	  ()
	(make-local-variable 'font-lock-background-mode)
	(setq font-lock-background-mode 'light)) ; Assume light bg
      (if (boundp 'font-lock-display-type)
	  ()
	(make-local-variable 'font-lock-display-type)
	(setq font-lock-display-type 'color)) ; Assume color

      ;; Create faces
      ;; Format: (FACE FOREGROUND BACKGROUND BOLD-P ITALIC-P UNDERLINE-P)
      (let* ((dark-bg (eq font-lock-background-mode 'dark))
	     (faces
	      (cond
	       ((memq font-lock-display-type '(mono monochrome))
		'((prolog-warning-face nil nil t t nil)
		  (prolog-builtin-face nil nil nil nil t)
		  (prolog-redo-face nil nil nil t nil)
		  (prolog-exit-face nil nil nil nil t)
		  (prolog-exception-face nil nil t t t)))
	       ((memq font-lock-display-type '(grayscale greyscale
							 grayshade greyshade))
		'((prolog-warning-face nil nil t t nil)
		  (prolog-builtin-face nil nil nil nil t)
		  (prolog-redo-face nil nil nil t nil)
		  (prolog-exit-face nil nil nil nil t)
		  (prolog-exception-face nil nil t t t)))
	       (dark-bg 		; dark colour background
		'((prolog-warning-face "red" nil t nil nil)
		  (prolog-builtin-face "LightSkyBlue" nil nil nil nil)
		  (prolog-redo-face "darkorchid" nil nil nil nil)
		  (prolog-exit-face "green" nil nil nil nil)
		  (prolog-exception-face "black" "Khaki" t nil nil)))
	       (t			; light colour background
		'((prolog-warning-face "red" nil t nil nil)
		  (prolog-builtin-face "Orchid" nil nil nil nil)
		  (prolog-redo-face "darkorchid" nil nil nil nil)
		  (prolog-exit-face "ForestGreen" nil nil nil nil)
		  (prolog-exception-face "black" "Khaki" t nil nil))))))
	(while faces
	  (if (fboundp 'font-lock-make-face)
	      (font-lock-make-face (car faces)) ; The preferred way
	    (make-face (nth 0 (car faces))))
	  (setq faces (cdr faces))))
      
      ;; Font Lock Patterns
      (let (
	    ;; "Native" Prolog patterns
	    (head-predicates
	     (list (format "^%s" prolog-atom-regexp)
		   0 font-lock-function-name-face))
	    (variables
	     '("\\<\\([_A-Z][a-zA-Z0-9_]*\\)"
	       1 font-lock-variable-name-face))
	    (important-elements
	     '("[][}{!;|]\\|\\*?->" 0 font-lock-keyword-face))
	    (predspecs			; module:predicate/cardinality
	     (list (format "\\<\\(%s:\\|\\)%s/[0-9]+"
			   prolog-atom-regexp prolog-atom-regexp)
		   0 font-lock-function-name-face 'prepend))
	    (keywords			; directives (queries)
	     (list (concat "^[?:]- *\\("
			   (prolog-make-keywords-regexp prolog-keywords)
			   "\\)\\>")
		   1 'prolog-builtin-face))
	    (quoted_atom (list prolog-quoted-atom-regexp
			       1 'font-lock-string-face 'append))
	    (string (list prolog-string-regexp
			  1 'font-lock-string-face 'append))
	    ;; SICStus specific patterns
	    (sicstus-object-methods
	     (list
	      'prolog-font-lock-object-matcher
	      1 font-lock-function-name-face))
	    ;; Inferior mode specific patterns
	    (prompt
	     (list prolog-prompt-regexp 0 'font-lock-keyword-face))
	    (trace-exit
	     (cond
	      ((eq prolog-system 'sicstus)
	       '("[ \t]*[0-9]+[ \t]+[0-9]+[ \t]*\\(Exit\\):"
		 1 'prolog-exit-face))
	      ((eq prolog-system 'swi)
	       '("[ \t]*\\(Exit\\):[ \t]*([ \t0-9]*)" 1 'prolog-exit-face))
	      (t nil)))
	    (trace-fail
	     (cond
	      ((eq prolog-system 'sicstus)
	       '("[ \t]*[0-9]+[ \t]+[0-9]+[ \t]*\\(Fail\\):"
		 1 'prolog-warning-face))
	      ((eq prolog-system 'swi)
	       '("[ \t]*\\(Fail\\):[ \t]*([ \t0-9]*)" 1 'prolog-warning-face))
	      (t nil)))
	    (trace-redo
	     (cond
	      ((eq prolog-system 'sicstus)
	       '("[ \t]*[0-9]+[ \t]+[0-9]+[ \t]*\\(Redo\\):"
		 1 'prolog-redo-face))
	      ((eq prolog-system 'swi)
	       '("[ \t]*\\(Redo\\):[ \t]*([ \t0-9]*)" 1 'prolog-redo-face))
	      (t nil)))
	    (trace-call
	     (cond
	      ((eq prolog-system 'sicstus)
	       '("[ \t]*[0-9]+[ \t]+[0-9]+[ \t]*\\(Call\\):"
		 1 font-lock-function-name-face))
	      ((eq prolog-system 'swi)
	       '("[ \t]*\\(Call\\):[ \t]*([ \t0-9]*)" 1 font-lock-function-name-face))
	      (t nil)))
	    (trace-exception
	     (cond
	      ((eq prolog-system 'sicstus)
	       '("[ \t]*[0-9]+[ \t]+[0-9]+[ \t]*\\(Exception\\):"
		 1 'prolog-exception-face))
	      ((eq prolog-system 'swi)
	       '("[ \t]*\\(Exception\\):[ \t]*([ \t0-9]*)"
		 1 'prolog-exception-face))
	      (t nil)))
	    (error-message-identifier
	     (cond
	      ((eq prolog-system 'sicstus)
	       '("{\\([A-Z]* ?ERROR:\\)" 1 'prolog-exception-face prepend))
	      ((eq prolog-system 'swi)
	       '("^[[]\\(WARNING:\\)" 1 'prolog-builtin-face prepend))
	      (t nil)))
	    (error-whole-messages
	     (cond
	      ((eq prolog-system 'sicstus)
	       '("{\\([A-Z]* ?ERROR:.*\\)}[ \t]*$"
		 1 font-lock-comment-face append))
	      ((eq prolog-system 'swi)
	       '("^[[]WARNING:[^]]*[]]$" 0 font-lock-comment-face append))
	      (t nil)))
	    (error-warning-messages
	     ;; Mostly errors that SICStus asks the user about how to solve,
	     ;; such as "NAME CLASH:" for example.
	     (cond
	      ((eq prolog-system 'sicstus)
	       '("^[A-Z ]*[A-Z]+:" 0 'prolog-warning-face))
	      (t nil)))
	    (warning-messages
	     (cond
	      ((eq prolog-system 'sicstus)
	       '("\\({ ?\\(Warning\\|WARNING\\) ?:.*}\\)[ \t]*$" 
		 2 'prolog-warning-face prepend))
	      (t nil))))

	;; Make font lock lists
	(defvar prolog-font-lock-keywords
	  (delete nil
		  (append (list head-predicates
				quoted_atom
				string
				variables
				important-elements
				predspecs
				keywords)
			  (if (eq prolog-system 'sicstus)
			      (list sicstus-object-methods)
			    ()))))
	(defvar inferior-prolog-font-lock-keywords
	  (delete nil
		  (list prompt
			error-message-identifier
			error-whole-messages
			error-warning-messages
			warning-messages
			predspecs
			trace-exit
			trace-fail
			trace-redo
			trace-call
			trace-exception)))
	(defvar prolog-compilation-font-lock-keywords
	  (delete nil
		  (list error-message-identifier
			error-whole-messages
			error-warning-messages
			warning-messages
			predspecs)))
	;; Add font lock lists to `font-lock-defaults-alist'
	(if (boundp 'font-lock-defaults-alist)
	    (progn
	      (if (not (assoc 'prolog-mode font-lock-defaults-alist))
		  (setq font-lock-defaults-alist
			(cons
			 '(prolog-mode . (prolog-font-lock-keywords
					  nil nil ((?_ . "w"))))
			 font-lock-defaults-alist)))
	      (if (not (assoc 'inferior-prolog-mode font-lock-defaults-alist))
		  (setq font-lock-defaults-alist
			(cons
			 '(inferior-prolog-mode . (inferior-prolog-font-lock-keywords
						   nil nil ((?_ . "w"))))
			 font-lock-defaults-alist))))))))


;;-------------------------------------------------------------------
;; Indentation stuff
;;-------------------------------------------------------------------

(defun prolog-indent-line (&optional whole-exp)
  "Indent current line as Prolog code.
With argument, indent any additional lines of the same clause
rigidly along with this one (not yet)."
  (interactive "p")
  (let ((indent (prolog-indent-level))
	(pos (- (point-max) (point))) beg)
    (beginning-of-line)
    (setq beg (point))
    (skip-chars-forward " \t")
    (if (zerop (- indent (current-column)))
	nil
      (delete-region beg (point))
      (indent-to indent))
    (if (> (- (point-max) pos) (point))
	(goto-char (- (point-max) pos)))
    
    ;; Align comments
    (if prolog-align-comments-flag
	(save-excursion
	  (prolog-goto-comment-column t)))

    (prolog-insert-spaces-after-paren)

    ))

(defun prolog-comment-indent ()
  "Compute prolog comment indentation."
  (cond ((looking-at "%%%") (prolog-indentation-level-of-line))
	((looking-at "%%") (prolog-indent-level))
	(t
	 (save-excursion
	   (skip-chars-backward " \t")
	   ;; Insert one space at least, except at left margin.
	   (max (+ (current-column) (if (bolp) 0 1))
		comment-column)))
	))

(defun prolog-indent-level ()
  "Compute prolog indentation level."
  (save-excursion
    (beginning-of-line)
    (let ((totbal (prolog-region-paren-balance
		   (prolog-clause-start t) (point)))
	  (oldpoint (point)))
      (skip-chars-forward " \t")
      (cond
       ((looking-at "%%%") (prolog-indentation-level-of-line))
					;Large comment starts
       ((looking-at "%[^%]") comment-column) ;Small comment starts
       ((bobp) 0)			;Beginning of buffer

       ;; If we found '}' then we must check if it's the
       ;; end of an object declaration or something else.
       ((and (looking-at "}") 
	     (save-excursion
	       (forward-char 1)
	       ;; Goto to matching {
	       (if prolog-use-prolog-tokenizer-flag
		   (prolog-backward-list)
		 (backward-list))
	       (skip-chars-backward " \t")
	       (backward-char 2)
	       (looking-at "::")))
	;; It was an object
	(if prolog-object-end-to-0-flag
	    0
	  prolog-indent-width)) 

       ;;End of /* */ comment
       ((looking-at "\\*/")                  
	(save-excursion
	  (prolog-find-start-of-mline-comment)
	  (skip-chars-backward " \t")
	  (- (current-column) 2)))

       ;; Here we check if the current line is within a /* */ pair
       ((and (looking-at "[^%/]")
	     (eq (prolog-in-string-or-comment) 'cmt)) 
	(if prolog-indent-mline-comments-flag
	    (prolog-find-start-of-mline-comment)
	  ;; Same as before
	  (prolog-indentation-level-of-line)))

       (t
	(let ((empty t) ind linebal)
	  ;; See previous indentation
	  (while empty
	    (forward-line -1)
	    (beginning-of-line)
	    (if (= (point) (point-min))
		(setq empty nil)
	      (skip-chars-forward " \t")
	      (if (not (or (prolog-in-string-or-comment)
			   (looking-at "%") 
			   (looking-at "\n")))
		  (setq empty nil))))

	  ;; Store this line's indentation
	  (if (= (point) (point-min))
	      (setq ind 0)		;Beginning of buffer
	    (setq ind (current-column))) ;Beginning of clause

	  ;; Compute the balance of the line
	  (setq linebal (prolog-paren-balance))

	  (if (< linebal 0)
	      (progn
		;; Add 'indent-level' mode to find-unmatched-paren instead?
		(end-of-line)
		(setq ind (prolog-find-indent-of-matching-paren))))

	  (beginning-of-line)

	  ;; Check if the line ends with ":-", ".", ":: {", "}" (might be
	  ;; unnecessary), "&" or ")" (The last four concerns SICStus objects)
	  (cond
	   ;; If the last char of the line is a '&' then set the indent level
	   ;; to prolog-indent-width (used in SICStus objects)
	   ((and (eq prolog-system 'sicstus) 
		 (looking-at ".+&[ \t]*\\(%.*\\|\\)$"))
	    (setq ind prolog-indent-width))

	   ;; Increase indentation if the previous line was the head of a rule
	   ;; and does not contain a '.'
	   ((looking-at (format ".*%s[^\\.]*[ \t]*\\(%%.*\\|\\)$" 
				prolog-head-delimiter))
	    (let (headindent)
	      (if (< (prolog-paren-balance) 0)
		  (save-excursion
		    (end-of-line)
		    (setq headindent (prolog-find-indent-of-matching-paren)))
		(setq headindent (prolog-indentation-level-of-line)))
	      (setq ind (+ headindent prolog-indent-width))))

	   ;; The previous line was the head of an object
	   ((looking-at ".+ *::.*{[ \t]*$")
	    (setq ind prolog-indent-width))

	   ;; If a '.' is found at the end of the previous line, then decrease
	   ;; the indentation. (The \\(%.*\\|\\) part of the regexp is for
	   ;; comments at the end of the line)
	   ((and (looking-at "^.+\\.[ \t]*\\(%.*\\|\\)$") 
		 ;; Make sure that the '.' found is not in a comment or string
		 (save-excursion
		   (end-of-line)
		   (re-search-backward "\\.[ \t]*\\(%.*\\|\\)$" (point-min))
		   (not (prolog-in-string-or-comment))))
	    (setq ind 0))

	   )

	  ;; If the last non comment char is a ',' or left paren or a left-
	  ;; indent-regexp then indent to open parenthesis level
	  (if (and
	       (> totbal 0)
	       ;; SICStus objects have special syntax rules if point is
	       ;; not inside additional parens (objects are defined
	       ;; within {...})
	       (not (and (eq prolog-system 'sicstus)
			 (= totbal 1)
			 (prolog-in-object))))
	      (if (looking-at
		   ;;(format "[^%%\n]*\\(,\\|%s\\|%s\\)[ \t]*\\(%%.*\\|\\)$" 
		   (format "\\(%s\\|%s\\|[^\n\\'\\\"%%]\\)*\\(,\\|%s\\|%s\\)\[ \t]*\\(%%.*\\|\\)$" 
			   prolog-quoted-atom-regexp prolog-string-regexp
			   prolog-left-paren prolog-left-indent-regexp))
		  (progn
		    (goto-char oldpoint)
		    (setq ind (prolog-find-unmatched-paren 'termdependent)))
		(goto-char oldpoint)
		(setq ind (prolog-find-unmatched-paren nil))
		))
	  

	  ;; Return the indentation level
	  ind
	  ))))))

(defun prolog-find-indent-of-matching-paren ()
  "Find the indentation level based on the matching parenthesis.
Indentation level is set to the one the point is after when the function is
called."
  (save-excursion
    ;; Go to the matching paren
    (if prolog-use-prolog-tokenizer-flag
	(prolog-backward-list)
      (backward-list))

    ;; If this was the first paren on the line then return this line's
    ;; indentation level
    (if (prolog-paren-is-the-first-on-line-p)
	(prolog-indentation-level-of-line)
      ;; It was not the first one
      (progn
	 ;; Find the next paren
	 (prolog-goto-next-paren 0)

	 ;; If this paren is a left one then use its column as indent level,
	 ;; if not then recurse this function
	 (if (looking-at prolog-left-paren)
	     (+ (current-column) 1)
	   (progn
	      (forward-char 1)
	      (prolog-find-indent-of-matching-paren)))
	 ))
    ))

(defun prolog-indentation-level-of-line ()
  "Return the indentation level of the current line."
  (save-excursion
    (beginning-of-line)
    (skip-chars-forward " \t")
    (current-column)))

(defun prolog-first-pos-on-line ()
  "Return the first position on the current line."
  (save-excursion
    (beginning-of-line)
    (point)))

(defun prolog-paren-is-the-first-on-line-p ()
  "Return t if the parenthesis under the point is the first one on the line.
Return nil otherwise.
Note: does not check if the point is actually at a parenthesis!"
  (save-excursion
    (let ((begofline (prolog-first-pos-on-line)))
      (if (= begofline (point))
	  t
	(if (prolog-goto-next-paren begofline)
	    nil
	  t)))))

(defun prolog-find-unmatched-paren (&optional mode)
  "Return the column of the last unmatched left parenthesis.
If MODE is `skipwhite' then any white space after the parenthesis is added to
the answer.
If MODE is `plusone' then the parenthesis' column +1 is returned.
If MODE is `termdependent' then if the unmatched parenthesis is part of
a compound term the function will work as `skipwhite', otherwise
it will return the column paren plus the value of `prolog-paren-indent'.
If MODE is nil or not set then the parenthesis' exact column is returned."
  (save-excursion
    ;; If the next paren we find is a left one we're finished, if it's
    ;; a right one then we go back one step and recurse
    (prolog-goto-next-paren 0)

    (let ((roundparen (looking-at "(")))
      (if (looking-at prolog-left-paren)
	  (let ((notpartofterm 
		 (save-excursion
		   (backward-char 1)
		   (looking-at "[ \t]"))))
	    (if (eq mode nil)
		(current-column)
	      (if (and roundparen
		       (eq mode 'termdependent) 
		       notpartofterm)
		  (+ (current-column) prolog-paren-indent)
		(forward-char 1)
		(if (or (eq mode 'skipwhite) (eq mode 'termdependent) )
		    (skip-chars-forward " \t"))
		(current-column))))
	;; Not looking at left paren
	(progn
	  (forward-char 1)
	  ;; Go to the matching paren. When we get there we have a total
	  ;; balance of 0.
	  (if prolog-use-prolog-tokenizer-flag
	      (prolog-backward-list)
	    (backward-list))
	  (prolog-find-unmatched-paren mode)))
      )))


(defun prolog-paren-balance ()
  "Return the parenthesis balance of the current line.
A return value of n means n more left parentheses than right ones."
  (save-excursion
    (end-of-line)
    (prolog-region-paren-balance (prolog-first-pos-on-line) (point))))

(defun prolog-region-paren-balance (start end)
  "Return the summed parenthesis balance in the region.
The region is limited by START and END positions."
  (save-excursion
    (let ((state (if prolog-use-prolog-tokenizer-flag
		     (prolog-tokenize start end)
		   (parse-partial-sexp start end))))
      (nth 0 state))))

(defun prolog-goto-next-paren (limit-pos)
  "Move the point to the next parenthesis earlier in the buffer.
Return t if a match was found before LIMIT-POS.  Return nil otherwise."
  (let (retval)
    (setq retval (re-search-backward
		  (concat prolog-left-paren "\\|" prolog-right-paren)
		  limit-pos t))

    ;; If a match was found but it was in a string or comment, then recurse
    (if (and retval (prolog-in-string-or-comment))
	(prolog-goto-next-paren limit-pos)
      retval)
    ))

(defun prolog-in-string-or-comment ()
  "Check whether string, atom, or comment is under current point.
Return:
 `txt' if the point is in a string, atom, or character code expression
 `cmt' if the point is in a comment
 nil otherwise."
  (save-excursion
    ;; NB! (point-min) is a safe point but it is inefficient to start
    ;; at the top of the buffer all the time!!!
    (let ((state (if prolog-use-prolog-tokenizer-flag
		     (prolog-tokenize (prolog-clause-start) (point))
		   (parse-partial-sexp (prolog-clause-start) (point)))))
      (cond
       ((nth 3 state) 'txt) ; String
       ((nth 4 state) 'cmt) ; Comment
       (t
	(cond
	 ((looking-at "%") 'cmt) ; Start of a comment
	 ((looking-at "/\\*") 'cmt) ; Start of a comment
	 ;;((looking-at "\'") 'txt) ; Start of an atom
	 ;;((looking-at "\"") 'txt) ; Start of a string
	 (t nil)
	 ))))
    ))

(defun prolog-find-start-of-mline-comment ()
  "Return the start column of a /* */ comment.
This assumes that the point is inside a comment."
  (re-search-backward "/\\*" (point-min) t)
  (forward-char 2)
  (skip-chars-forward " \t")
  (current-column))

(defun prolog-insert-spaces-after-paren ()
  "Insert spaces after the opening parenthesis.
Spaces are inserted if the parenthesis is the first non-white character
on the line."
  (save-excursion
    (let (incr level)
      (beginning-of-line)
      (skip-chars-forward " \t")
      (if (looking-at (format "(\\|%s" prolog-left-indent-regexp))
	  (progn
	    ;; Treat "( If -> " lines specially.
	    (if (looking-at "(.*->")
		(setq incr 2)
	      (setq incr prolog-paren-indent))
	    
	    ;; Goto the first char after the matched text
	    (re-search-forward (format "(\\|%s" prolog-left-indent-regexp)
			       (point-max) t)
	    (goto-char (match-end 0))
	    
	    (setq level (+ (prolog-find-unmatched-paren) incr))
	    
	    ;; Remove old white space
	    (let ((start (point)))
	      (skip-chars-forward " \t")
	      (delete-region start (point)))
	    
	    (indent-to level)
	    ))))
  (skip-chars-forward " \t"))


;;-------------------------------------------------------------------
;; The tokenizer
;;-------------------------------------------------------------------

(defconst prolog-tokenize-searchkey
  (concat "[0-9]+'"
	  "\\|"
	  "['\"]"
	  "\\|"
	  prolog-left-paren
	  "\\|"
	  prolog-right-paren
	  "\\|"
	  "%"
	  "\\|"
	  "/\\*"
	  ))

(defun prolog-tokenize ( start end &optional stopcond )
  "Tokenizes a region of prolog code.
START and END are positions.
STOPCOND decides the stop condition of the parsing. Valid values
are 'zerodepth which stops the parsing at the first right parenthesis
where the parenthesis depth is zero, and nil.

The function returns a list with the following information:
 0. parenthesis depth 
 3. 'atm if END is inside an atom
    'str if END is inside a string
    'chr if END is in a character code expression (0'x)
    nil otherwise
 4. non-nil if END is inside a comment
 5. end position (always equal to END if STOPCOND is nil)
The rest of the elements are undefined."
  (save-excursion
    (let* ((end2 (+ end 1))
	   tmpstr
	   oldp
	   (depth 0)
	   (quoted nil)
	   inside_cmt
	   (endpos end2)
	   )
      
      (goto-char start)
      (while (and
	      (re-search-forward prolog-tokenize-searchkey end2 t)
	      (< (point) end2))
	(progn
	  (setq tmpstr (match-string 0))
	  (setq oldp (point))
	  (goto-char (match-beginning 0))
	  (cond
	   ;; 0'char
	   ((looking-at "0'")
	    (setq oldp (1+ (match-end 0)))
	    (if (> oldp end) 
		(setq quoted 'chr)))
	   
	   ;; base'number
	   ((looking-at "[0-9]+'")
	    (goto-char (match-end 0))
	    (skip-chars-forward "0-9a-za-Z")
	    (setq oldp (point)))

	   ;; Paren stuff
	   ((looking-at prolog-left-paren)
	    (setq depth (1+ depth)))

	   ((looking-at prolog-right-paren)
	    (setq depth (1- depth))
	    (if (and
		 (eq stopcond 'zerodepth)
		 (= depth 0))
		(progn
		  (setq endpos (point))
		  (setq oldp end2))))

	   ;; Comment stuff
	   ((looking-at comment-start)
	    (end-of-line)
	    (if (>= (point) end2)
		(progn
		  (setq inside_cmt t)
		  (setq oldp end2))
	      (setq oldp (point))))

	   ((looking-at "/\\*")
	    (if (re-search-forward "\\*/" end2 'limit)
		(setq oldp (point))
	      (setq inside_cmt t)
	      (setq oldp end2)))

	   ;; Atoms and strings
	   ((looking-at "'")
	    ;; Find end of atom
	    (if (re-search-forward "[^\\]'" end2 'limit)
		()
	      (setq quoted 'atm))
	    (setq oldp (point)))

	   ((looking-at "\"")
	    ;; Find end of string
	    (if (re-search-forward "[^\\]\"" end2 'limit)
		()
	      (setq quoted 'str))
	    (setq oldp (point)))
	   )
	  (goto-char oldp)
	  ))				; End of while
      
      ;; Create return list
      (list depth nil nil quoted inside_cmt endpos)
      )))


;;-------------------------------------------------------------------
;; Online help
;;-------------------------------------------------------------------

(defvar prolog-help-function
  (cond
   ((eq prolog-system 'sicstus) 'prolog-help-sicstus)
   ((eq prolog-system 'swi) 'prolog-help-swi)
   (t 'prolog-help-swi))
  "The name of the function which finds help on a predicate.")

;; Inspired by `word-help.el'
(defun prolog-help-sicstus (predicate)
  (let ((buffer (current-buffer)))
    (require 'info)
    (pop-to-buffer nil)
    (Info-goto-node "(sicstus)Built-In Index")
    (if (not (re-search-forward
	      (concat "^\\* " (regexp-quote predicate) " */") nil t))
	(error (format "Help on predicate `%s' not found." predicate)))
    (re-search-backward "[^ /]" nil t)
    (Info-follow-nearest-node)
    (re-search-forward (concat "^`" (regexp-quote predicate)) nil t)
    (beginning-of-line)
    (recenter 0)
    (pop-to-buffer buffer)))

(defun prolog-help-swi (predicate)
  (ensure-prolog-process)
  (process-send-string "prolog" (concat "help(" predicate ").\n"))
  (display-buffer "*prolog*"))

(defun prolog-help-on-predicate ()
  "Invoke online help on the atom under cursor."
  (interactive)
  (let* ((case-fold-search nil)
	 (regexp-1 prolog-atom-regexp)
	 (regexp-2 "[-<>=*+@$:\\]+")
	 (regexp (format "\\(\\<%s\\>\\|%s\\)" regexp-1 regexp-2))
	 (regexp-1-point (concat regexp-1 "\\="))
	 (regexp-2-point (concat regexp-2 "\\="))
	 word
	 predicate
	 point)
    (save-excursion
      (if (or (looking-at prolog-atom-char-regexp)
	      (and (/= (point) (point-min)) ; Don't go past BOB
		   (save-excursion
		     (backward-char 1)
		     (looking-at prolog-atom-char-regexp))))
	  (re-search-backward "\\<" nil t))
      (setq point (if (re-search-forward regexp nil t)
		      (match-end 0)
		    (point)))
      (re-search-backward regexp nil t)
      (if (looking-at regexp-1)
	  (while (re-search-backward regexp-1-point nil t))
	(while (re-search-backward regexp-2-point nil t)))
      (setq word (and (not (eq (point) point))
		      (buffer-substring-no-properties (point) point))))
    (setq predicate (read-from-minibuffer
		     (format "Help on predicate%s: "
			     (if word
				 (concat " (default " word ")")
			       ""))))
    (if (string= predicate "")
	(setq predicate word))
    (funcall prolog-help-function predicate)))

(defun prolog-help-apropos (string)
  "Find Prolog apropos on given STRING.
This function is only available when `prolog-system' is set to `swi'."
  (interactive "sApropos: ")
  (cond
   ((eq prolog-system 'swi)
    (ensure-prolog-process)
    (process-send-string "prolog" (concat "apropos(" string ").\n")))
   (t
    (error "Sorry, no Prolog apropos available for your system."))))


;;-------------------------------------------------------------------
;; Miscellaneous functions
;;-------------------------------------------------------------------

;; For Windows. Change backslash to slash. SICStus handles either
;; path separator but backslash must be doubled, therefore use slash.
(defun prolog-bsts (string)
  "Change backslashes to slashes in STRING."
  (let ((str1 (copy-sequence string))
	(len (length string))
	(i 0))
    (while (< i len)
      (if (char-equal (aref str1 i) ?\\)
	  (aset str1 i ?/))
      (setq i (1+ i)))
    str1))

;; For backward compatibility. Stolen from custom.el.
(or (fboundp 'match-string)
    ;; Introduced in Emacs 19.29.
    (defun match-string (num &optional string)
  "Return string of text matched by last search.
NUM specifies which parenthesized expression in the last regexp.
 Value is nil if NUMth pair didn't match, or there were less than NUM pairs.
Zero means the entire text matched by the whole regexp or whole string.
STRING should be given if the last search was by `string-match' on STRING."
  (if (match-beginning num)
      (if string
	  (substring string (match-beginning num) (match-end num))
	(buffer-substring (match-beginning num) (match-end num))))))

(defun prolog-pred-start ()
  "Return the starting point of the first clause of the current predicate."
  (save-excursion
    (goto-char (prolog-clause-start))
    ;; Find first clause, unless it was a directive
    (if (and
	 (not (looking-at ":-"))
	 (not (looking-at "[ \t]*[%/]")) ; Comment
	 )
	(let (predname (op (point)))
	  (skip-chars-forward "^[ (\\.]")
	  (setq predname (buffer-substring op (point)))
	  (beginning-of-line)
	  (while (re-search-backward (concat "^" predname "[ (\\.]") nil t))))
      (point)))

(defun prolog-pred-end ()
  "Return the position at the end of the last clause of the current predicate."
  (save-excursion
    (goto-char (prolog-clause-end))
    (goto-char (prolog-clause-start))
    (let (predname (op (point)))
      (if (looking-at ":-")
	  ;; This was a directive
	  (goto-char (prolog-clause-end))
	;; It was not a directive, find the last clause
	(skip-chars-forward "^[ (\\.]")
	(setq predname (buffer-substring op (point)))
	(beginning-of-line)
	(while (re-search-forward (concat "^" predname "[ (\\.]") nil t)
	  (goto-char (prolog-clause-end))))
      (point))))

(defun prolog-clause-start ( &optional notallowmethods )
  "Return the position at the start of the head of the current clause.
If NOTALLOWMETHODS is non-nil then do not match on methods in
objects (relevent only if 'prolog-system' is set to sicstus)."
  (save-excursion
    (let ((notdone t) (retval (point-min)))
      (end-of-line)
      (while (and 
	      notdone 
	      ;; Search for a head or a fact
	      (re-search-backward
	       (if (and (not notallowmethods)
			(eq prolog-system 'sicstus)
			(prolog-in-object))
		   ;; If in object, then find method start.
		   "^\\([ \t]+[a-z$].*\\(:-\\|&\\)\\|[ \t]+[a-z$].*:: {\\)" 
		 (format "^\\(\\([a-z$].*\\(%s\\|\\.\\|,\\)\\)\\|:-\\)" 
			 prolog-head-delimiter))
	       (point-min) t))
	(if (>= (prolog-paren-balance) 0)
	    ;; Start of clause found
	    (progn
	      (setq retval (point))
	      (setq notdone nil))))
      retval)))

(defun prolog-clause-end ( &optional notallowmethods )
  "Return the position at the end of the current clause.
If NOTALLOWMETHODS is non-nil then do not match on methods in
objects (relevent only if 'prolog-system' is set to sicstus)."
  (save-excursion
    (beginning-of-line)			; Necessary since we use "^...." 
					; for the search
    (if (re-search-forward 
	 (if (and (not notallowmethods)
		  (eq prolog-system 'sicstus)
		  (prolog-in-object))
	     "^\\(.*&[ \t]*\\(\\|%%.*\\)$\\|[ \t]*}\\)"
	   (format "^\\(%s\\|%s\\|[^\n\\'\\\"%%]\\)*\\.[ \t]*\\(\\|%%.*\\)$"
		   prolog-quoted-atom-regexp prolog-string-regexp))
	 (point-max) t)
	(point)
      (error "End of clause not found!"))))

(defun prolog-in-object ()
  "Return t if the point is inside a SICStus object definition."
  ;; Return t if the last line that starts with a character
  ;; that is neither white space nor a comment start
  (save-excursion
    (beginning-of-line)
    (if (looking-at ".*::[ \t]*{")
	t
      (re-search-backward "^[^ \t\n%/\\*]" (point-min) 'limit)
      (looking-at ".*::[ \t]*{"))))

(defun prolog-electric-newline ()
  "Insert newline and indent the next line."
  (interactive)
  ;; Remove trailing white space
  (let (p)
    (skip-chars-forward " \t")    
    (setq p (point))
    (skip-chars-backward " \t")
    (delete-region (point) p))
  (newline-and-indent))

(defun prolog-forward-list ()
  "Move the point to the matching right parenthesis."
  (interactive)
  (if prolog-use-prolog-tokenizer-flag
      (let ((state (prolog-tokenize (point) (point-max) 'zerodepth)))
	(goto-char (+ (nth 5 state) 1)))
    (forward-list)))

;; NB: This could be done more efficiently!
(defun prolog-backward-list ()
  "Move the point to the matching left parenthesis."
  (interactive)
  (if prolog-use-prolog-tokenizer-flag
      (let ((bal 0)
	    (paren-regexp (concat prolog-left-paren "\\|" prolog-right-paren))
	    (notdone t))
	(while (and notdone (re-search-backward paren-regexp nil t))
	  (cond
	   ((looking-at prolog-left-paren)
	    (if (not (prolog-in-string-or-comment))
		(setq bal (1+ bal)))
	    (if (= bal 0)
		(setq notdone nil)))
	   ((looking-at prolog-right-paren)
	    (if (not (prolog-in-string-or-comment))
		(setq bal (1- bal))))
	   )))
    (backward-list)))

(defun prolog-goto-clause-start ()
  (interactive)
  (goto-char (prolog-clause-start)))

(defun prolog-goto-clause-end ()
  (interactive)
  (goto-char (prolog-clause-end)))

(defun prolog-goto-pred-start ()
  (interactive)
  (let ((op (point)))
    (goto-char (prolog-pred-start))
    (if (= op (point))
	(progn
	  (backward-char 1)
	  (prolog-goto-pred-start)))))

(defun prolog-goto-pred-end ()
  (interactive)
  (let ((op (point)))
    (goto-char (prolog-pred-end))
    (if (= op (point))
	(progn
	  (forward-line 1)
	  (prolog-goto-pred-end)))))

(defun prolog-insert-predicate-name ()
  "Insert the name of the current predicate."
  (interactive)
  (let (predname)
    (save-excursion
      (goto-char (prolog-clause-start))
      (let ((op (point)))
	(skip-chars-forward "^[ (\\.]")
	(setq predname (buffer-substring op (point)))))
    (insert predname)))

(defun prolog-insert-next-clause ()
  "Insert newline and the name of the current clause."
  (interactive)
  (insert "\n")
  (prolog-insert-predicate-name))

(defun prolog-insert-module-modeline ()
  "Insert a modeline for module specification.
This line should be first in the buffer.
The module name should be written manually just before the semi-colon."
  (interactive)
  (insert "%%% -*- Module: ; -*-\n")
  (backward-char 6))

(defun prolog-uncomment-region ( start end )
  "Uncomment the region."
  (interactive "r")
  (comment-region start end -1))

(defun prolog-goto-comment-column (&optional nocreate)
  "Move comments on the current line to the correct position.
If NOCREATE is nil (or omitted) and there is no comment on the line, then
a new comment is created."
  (interactive)
  (beginning-of-line)
  (if (or (not nocreate)
	  (and
	   (re-search-forward 
	    (format "^\\(\\(%s\\|%s\\|[^\n\\'\"%%]\\)*\\)%% *" 
		    prolog-quoted-atom-regexp prolog-string-regexp)
	    (save-excursion (end-of-line) (point)) 'limit)
	   (progn
	     (goto-char (match-beginning 0))
	     (not (eq (prolog-in-string-or-comment) 'txt)))))
      (indent-for-comment)))

(defun prolog-indent-predicate ()
  "*Indent the current predicate."
  (interactive)
  (indent-region (prolog-pred-start) (prolog-pred-end) nil))

(defun prolog-indent-buffer ()
  "*Indent the entire buffer."
  (interactive)
  (indent-region (point-min) (point-max) nil))

(defun prolog-mark-clause ()
  "Put mark at the end of this clause and move point to the beginning."
  (interactive)
  (let ((pos (point)))
    (goto-char (prolog-clause-end))
    (forward-line 1)
    (beginning-of-line)
    (set-mark (point))
    (goto-char pos)
    (goto-char (prolog-clause-start))))

(defun prolog-mark-predicate ()
  "Put mark at the end of this predicate and move point to the beginning."
  (interactive)
  (let (pos)
    (goto-char (prolog-pred-end))
    (setq pos (point))
    (forward-line 1)
    (beginning-of-line)
    (set-mark (point))
    (goto-char pos)
    (goto-char (prolog-pred-start))))

;; Stolen from `cc-mode.el':
(defun prolog-electric-delete (arg)
  "Deletes preceding character or whitespace.
If `prolog-hungry-delete-key-flag' is non-nil, then all preceding whitespace is
consumed.  If however an ARG is supplied, or `prolog-hungry-delete-key-flag' is
nil, or point is inside a literal then the function in the variable
`backward-delete-char' is called."
  (interactive "P")
  (if (or (not prolog-hungry-delete-key-flag)
	  arg
	  (eq (prolog-in-string-or-comment) 'txt))
      (funcall 'backward-delete-char (prefix-numeric-value arg))
    (let ((here (point)))
      (skip-chars-backward " \t\n")
      (if (/= (point) here)
	  (delete-region (point) here)
	(funcall 'backward-delete-char 1)
	))))


;;-------------------------------------------------------------------
;; Menu stuff (both for the editing buffer and for the inferior
;; prolog buffer)
;;-------------------------------------------------------------------

(defun prolog-menu ()
  (easy-menu-define prolog-menu-map (current-local-map) ""
		    (list "Prolog"
			  '("Consult"
			    ["Buffer" prolog-consult-buffer t]
			    ["Region" prolog-consult-region t]
			    ["Predicate" prolog-consult-predicate t])
			  '("Compile"
			    ["Buffer" prolog-compile-buffer t]
			    ["Region" prolog-compile-region t]
			    ["Predicate" prolog-compile-predicate t])
			  '("Debug"
			   ["on" prolog-debug-on t]
			   ["off" prolog-debug-off t])
			  '("Trace"
			   ["on" prolog-trace-on t]
			   ["off" prolog-trace-off t])
			  ["Run Prolog" run-prolog t]
			  "-"
			  ["Comment region" comment-region t]
			  ["Uncomment region" prolog-uncomment-region t]
			  ["Goto comment column" indent-for-comment t]
			  "-"
			  '("Insert"
			   ["Predicate name" prolog-insert-predicate-name t]
			   ["Module modeline" prolog-insert-module-modeline t])
			  '("Goto"
			   ["Start of clause" prolog-goto-clause-start t]
			   ["End of clause" prolog-goto-clause-end t]
			   ["Start of predicate" prolog-goto-pred-start t]
			   ["End of predicate" prolog-goto-pred-end t])
			  '("Fontify"
			   ["Buffer" font-lock-fontify-buffer t])
			  '("Indent"
			   ["Line" prolog-indent-line t]
			   ["Region" indent-region t]
			   ["Predicate" prolog-indent-predicate t]
			   ["Buffer" prolog-indent-buffer t])
			  '("Mark"
			   ["Clause" prolog-mark-clause t]
			   ["Predicate" prolog-mark-predicate t]
			   ["Paragraph" mark-paragraph t])
			  (cond 
			   ((eq prolog-system 'swi)
			    '("Help"
			      ["On Predicate" prolog-help-on-predicate t]
			      ["Apropos" prolog-help-apropos t]))
			   ((eq prolog-system 'sicstus)
			    '("Help"
			      ["On Predicate" prolog-help-on-predicate t]))
			   (t
			    '("Help"
			      ["On Predicate" prolog-help-on-predicate t]))
			   )
			   )))

;; Menu for the inferior prolog buffer
(defun inferior-prolog-menu ()
  (easy-menu-define run-prolog-menu-map (current-local-map) ""
		    '("Prolog"
		      ("Debug"
		       ["on" prolog-debug-on t]
		       ["off" prolog-debug-off t])
		      ("Trace"
		       ["on" prolog-trace-on t]
		       ["off" prolog-trace-off t])
		      "-"
		      ["Interrupt Prolog" comint-interrupt-subjob t]
		      ["Stop Prolog" comint-stop-subjob t]
		      ["Continue Prolog" comint-continue-subjob t]
		      ["Quit Prolog" comint-quit-subjob t]
		      ["Kill Prolog" comint-kill-subjob t]
		      ["Send EOF" comint-send-eof t]
		      "-"
		      ["List Input History" comint-dynamic-list-input-ring t]
		      ["Previous Input" comint-previous-input t]
		      ["Next Input" comint-next-input t]
		      ["Previous Matching Current Input"
		       comint-previous-matching-input-from-input t]
		      ["Next Matching Current Input"
		       comint-next-matching-input-from-input t]
		      ["Previous Matching Input..."
		       comint-previous-matching-input t]
		      ["Next Matching Input..." comint-next-matching-input t]
		      ["Backward Matching Input..."
		       comint-backwards-matching-input t]
		      ["Forward Matching Input..."
		       comint-forwards-matching-input t]
		      ["Copy Old Input" comint-copy-old-input t]
		      ["Kill Current Input" comint-kill-input t]
		      ["Show Current Output Group" comint-show-output t]
		      ["Show Maximum Output" comint-show-maximum-output t]
		      ["Backward Output Group" comint-previous-prompt t]
		      ["Forward Output Group" comint-next-prompt t]
		      ["Kill Current Output Group" comint-kill-output t]
		      ["Kill Word" backward-kill-word t]
		      ["Beginning of Line" comint-bol t]
		      )))

(defun prolog-debug-on (&optional prefix)
  "Enable debugging.
When called with prefix argument, disable debugging instead."
  (interactive "P")
  (if prefix
      (prolog-debug-off)
    (prolog-process-insert-string (get-process "prolog")
				  prolog-debug-on-string)
    (process-send-string "prolog" prolog-debug-on-string)))

(defun prolog-debug-off ()
  "Disable debugging."
  (interactive)
  (prolog-process-insert-string (get-process "prolog")
				prolog-debug-off-string)
  (process-send-string "prolog" prolog-debug-off-string))

(defun prolog-trace-on (&optional prefix)
  "Enable tracing.
When called with prefix argument, disable tracing instead."
  (interactive "P")
  (if prefix
      (prolog-trace-off)
    (prolog-process-insert-string (get-process "prolog")
				  prolog-trace-on-string)
    (process-send-string "prolog" prolog-trace-on-string)))

(defun prolog-trace-off ()
  "Disable tracing."
  (interactive)
  (prolog-process-insert-string (get-process "prolog")
				prolog-trace-off-string)
  (process-send-string "prolog" prolog-trace-off-string))


(add-hook 'prolog-mode-hook 'prolog-menu)
(add-hook 'inferior-prolog-mode-hook 'inferior-prolog-menu)

(provide 'prolog)  

;;; prolog.el ends here

